/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

#if !defined (_SPRC_TYP_)
#define _SPRC_TYP_

#include "edc_typ.h"
#include "dci_typ.h"

class sprc_typ {	// search procedure for analysis of residue couplings
   public:
                sprc_typ(){ print_error("sprc_typ( ) constructor is disallowed"); }
                sprc_typ(int ac,char *av[],dch_typ *dca_heap)
		{ 
			argc=ac; Mode='M'; NEWP(argv,ac+3,char);
			for(Int4 i=0; i < argc; i++) argv[i]=av[i];
			mds_file=src_pdb=mds_ptrn=0;
			dch=dca_heap; dcaBGfile=0; Init();
		}
                ~sprc_typ( ){ Free(); }	
	Int4	run_mstarc(FILE *vsifp=0,char *pdb_id=0,FILE *dcfp=0,FILE *infp=0,
			char *mssd_nm=0,FILE *osfp=0,FILE *lgfp=0);
	void	SetBGfile(char *name){ assert(dcaBGfile==0); dcaBGfile=AllocString(name); }
	void	PassInMDS(char *mf,char *sp,char *mp=0){ mds_file=mf; src_pdb=sp; mds_ptrn=mp; }
	Int4	RunCMPSS(char str[],char PDBID[], char c, char mode, FILE *lgfp=0,
				set_typ SetUsd=0){
		   Int4 num_dc=0,values[5]; values[0]=0;
		   return RunSTARC(str,PDBID,c,values,0,mode,0,num_dc,0,0,lgfp,SetUsd,TRUE);
		}
   private:
        char    *mds_file,*src_pdb,*mds_ptrn; 	// don't free these up...freed in calling environment.
	void	Init();
	void	Free(){ if(dcaBGfile) free(dcaBGfile); free(argv); if(0) delete dch; }
	Int4	IsChainOkayPDB(char c, char PDBID[],pdb_typ PDB);
	Int4	RunSTARC(char str[],char PDBID[], char c, Int4 values[],char ColorCode,
		   char mode, FILE *vsifp=0,Int4 num_dc=0,FILE *dcfp=0, FILE *osfp=0,
		   FILE *lgfp=0,set_typ SetUsd=0,BooLean IgnoreAdj=FALSE);
	dch_typ	*dch;
	int	argc,argB;
	char	**argv,Mode;
	float	Dmax;
	char	*dcaBGfile;
	Int4	num_dc;
};

#endif

